"""
Contains the Typer app of the fundamend CLI tool.
"""

import typer
from rich.console import Console

app = typer.Typer(name="fundamend", help="CLI tool to work with XML files by BDEW", no_args_is_help=True)
err_console = Console(stderr=True)  # https://typer.tiangolo.com/tutorial/printing/#printing-to-standard-error



"""internal helper functions"""

from pathlib import Path

try:
    import sqlalchemy
    from sqlmodel import Session
except ImportError as import_error:
    import_error.msg += "; Did you install fundamend[sqlmodels] or did you try to import from fundamend.models instead?"
    # sqlmodel is only an optional dependency when fundamend is used to fill a database
    raise


def _execute_bare_sql(session: Session, path_to_sql_commands: Path) -> None:
    """
    Execute bare SQL from the path_to_sqlcommands in the given SQLAlchemy session.
    """

    with open(path_to_sql_commands, "r", encoding="utf-8") as sql_file:
        bare_sql = sql_file.read()

    bare_statements = bare_sql.split(";")

    for bare_statement in bare_statements:
        statement = bare_statement.strip()
        if statement:
            try:
                session.execute(sqlalchemy.text(statement))
            except sqlalchemy.exc.IntegrityError:
                if " UNIQUE " in bare_statement:
                    session.execute(sqlalchemy.text(bare_statement.replace(" UNIQUE ", " ")))
                else:
                    raise
    session.commit()



"""
Contains the command to convert XML files to JSON files.
"""

import re
from itertools import groupby
from pathlib import Path
from typing import Iterator, Literal

import typer
from typing_extensions import Annotated

from fundamend import AhbReader, Anwendungshandbuch, MessageImplementationGuide, MigReader
from fundamend.commands.app import app
from fundamend.sanitize import sanitize_ahb

FORMAT_AND_TYPE_REGEX = re.compile(r"^([A-Z]+)_(AHB|MIG)_(?:(Gas|Strom)_)?")


def _write_ahb_models_splitted(
    model: Anwendungshandbuch,
    ahb_dir: Path,
    *,
    compressed: bool = False,
) -> None:
    """Writes the given Anwendungshandbuch model to multiple JSON files, one for each Anwendungsfall."""
    ahb_dir.mkdir(parents=True, exist_ok=True)
    for anwendungsfall in model.anwendungsfaelle:
        json_file_path = ahb_dir / f"{anwendungsfall.pruefidentifikator}.json"
        with open(json_file_path, encoding="utf-8", mode="w") as outfile:
            outfile.write(anwendungsfall.model_dump_json(indent=None if compressed else 2))

    # Write meta file
    ahb_meta_file_path = ahb_dir / "meta.json"
    with open(ahb_meta_file_path, encoding="utf-8", mode="w") as outfile:
        outfile.write(model.model_dump_json(exclude={"anwendungsfaelle"}, indent=None if compressed else 2))


def _write_model_to_json_file(
    model: Anwendungshandbuch | MessageImplementationGuide,
    xml_file_path: Path,
    *,
    compressed: bool = False,
    split_ahb: bool = False,
) -> None:
    """Writes the given model to a JSON file at the specified path."""
    if split_ahb:
        if not isinstance(model, Anwendungshandbuch):
            raise ValueError("split_ahb can only be used with Anwendungshandbuch models")
        ahb_dir = xml_file_path.with_suffix("")
        _write_ahb_models_splitted(model, ahb_dir, compressed=compressed)
        typer.echo(f"Successfully converted {xml_file_path} to multiple JSON files in {ahb_dir}")
    else:
        json_file_path = xml_file_path.with_suffix(".json")
        with open(json_file_path, encoding="utf-8", mode="w") as outfile:
            outfile.write(model.model_dump_json(indent=None if compressed else 2))
        typer.echo(f"Successfully converted {xml_file_path} to JSON {json_file_path}")


def _convert_to_json_files(
    mig_xml_file_path: Path, ahb_xml_file_path: Path, sanitize: bool = False
) -> tuple[MessageImplementationGuide, Anwendungshandbuch]:
    """converts the given XML file to a JSON file and returns the path of the latter"""
    if not mig_xml_file_path.is_file():  # pragma: no cover
        raise ValueError(f"The given path {mig_xml_file_path.absolute()} is not a file")
    if not ahb_xml_file_path.is_file():  # pragma: no cover
        raise ValueError(f"The given path {ahb_xml_file_path.absolute()} is not a file")

    mig_model = MigReader(mig_xml_file_path).read()
    ahb_model = AhbReader(ahb_xml_file_path).read()

    # Do sanitization if requested
    if sanitize:
        sanitize_ahb(mig_model, ahb_model)

    return mig_model, ahb_model


def xml2json_dir_mode(
    xml_path: Path, sanitize: bool = False, compressed: bool = False, split_ahb: bool = False
) -> None:
    """
    Converts all XML files in the given directory to JSON files.
    The function expects to find pairs of MIG and AHB XML files in the directory.
    The XML file names must match the pattern `<FORMAT>_<AHB|MIG>_[<Gas|Strom>_]*.xml`.
    """

    def groupby_key(path_and_match: tuple[Path, re.Match[str] | None]) -> str:
        assert path_and_match[1] is not None
        return path_and_match[1].group(1) + (path_and_match[1].group(3) or "")

    def sort_key(path_and_match: tuple[Path, re.Match[str] | None]) -> str:
        assert path_and_match[1] is not None
        return groupby_key(path_and_match) + path_and_match[1].group(2)

    def xmls_and_matches() -> Iterator[tuple[Path, re.Match[str]]]:
        for _xml_path in xml_path.rglob("*.xml"):
            match = FORMAT_AND_TYPE_REGEX.match(_xml_path.name)
            if match is None:  # pragma: no cover
                raise ValueError("XML file name does not match expected format: " + str(_xml_path))
            yield _xml_path, match

    for _, _xmls_and_matches in groupby(sorted(xmls_and_matches(), key=sort_key), key=groupby_key):
        _xmls_and_matches_list = list(_xmls_and_matches)
        assert len(_xmls_and_matches_list) == 2, (
            "Expected exactly two XML files (AHB + MIG) for each format and powert type, but found: "
            f"{_xmls_and_matches_list}"
        )
        assert (
            _xmls_and_matches_list[0][1].group(2) == "AHB" and _xmls_and_matches_list[1][1].group(2) == "MIG"
        ), f"Expected AHB on first and a MIG on second position, but found: {_xmls_and_matches_list}"
        ahb_path = _xmls_and_matches_list[0][0]
        mig_path = _xmls_and_matches_list[1][0]
        mig, ahb = _convert_to_json_files(mig_path, ahb_path, sanitize=sanitize)
        _write_model_to_json_file(mig, mig_path.with_suffix(".json"), compressed=compressed)
        _write_model_to_json_file(ahb, ahb_path.with_suffix(".json"), compressed=compressed, split_ahb=split_ahb)


def xml2json_file_mode(
    xml_path: Path, sanitize: bool = False, compressed: bool = False, split_ahb: bool = False
) -> None:
    """
    Converts a single XML file to JSON.
    The function expects to find the corresponding AHB or MIG file in the same directory.

    The XML file names must match the pattern `<FORMAT>_<AHB|MIG>_[<Gas|Strom>_]*.xml`.
    """
    match = FORMAT_AND_TYPE_REGEX.match(xml_path.name)
    if match is None:  # pragma: no cover
        raise ValueError("XML file name does not match expected format: " + str(xml_path))
    match_type: Literal["MIG", "AHB"] = match.group(2)  # type: ignore[assignment]
    match_type_other: Literal["MIG", "AHB"] = "AHB" if match_type == "MIG" else "MIG"
    pattern_other = f"{match.group(1)}_{match_type_other}_"
    if match.group(3) is not None:
        pattern_other += f"{match.group(3)}_"
    pattern_other += "*.xml"

    other_matches = list(xml_path.parent.glob(pattern_other))
    if len(other_matches) == 0:  # pragma: no cover
        raise ValueError(
            f"No other XML file found in the same directory as {xml_path} matching pattern {pattern_other}"
        )
    if len(other_matches) > 1:  # pragma: no cover
        raise ValueError(
            f"Multiple other XML files found in the same directory as {xml_path} matching pattern "
            f"{pattern_other}: {other_matches}"
        )
    if match_type == "MIG":
        mig, ahb = _convert_to_json_files(xml_path, other_matches[0], sanitize=sanitize)
        _write_model_to_json_file(mig, xml_path.with_suffix(".json"), compressed=compressed)
    else:
        mig, ahb = _convert_to_json_files(other_matches[0], xml_path, sanitize=sanitize)
        _write_model_to_json_file(ahb, xml_path.with_suffix(".json"), compressed=compressed, split_ahb=split_ahb)


@app.command()
def xml2json(
    xml_path: Annotated[
        Path,
        typer.Option(
            ...,
            "--xml-path",
            "-p",
            exists=True,
            file_okay=True,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    sanitize: Annotated[
        bool,
        typer.Option(
            ...,
            "--sanitize",
            "-s",
            help="Sanitize the MIG or AHB before writing the resulting JSON. As of now, it does two things:\n"
            '1) Data elements or groups which are stated as "unused" in the MIG are missing in the AHB. '
            "The sanitization will add them to the AHB to enable easy parallel iteration over MIG and AHB. \n"
            "2) The five data elements C_C080 D_3036 model names. But in AHB there is only one D_3036 with "
            'description "Name". The sanitization will add four extra D_3036 data elements to prevent reading'
            "raster errors.",
        ),
    ] = False,
    compressed: Annotated[
        bool,
        typer.Option(
            ...,
            "--compressed",
            "-c",
            help="If set, the output JSON files will contain no whitespace outside of strings. If not set"
            " (default), the output JSON files will be pretty-printed with an indentation of one space.",
        ),
    ] = False,
    split_ahb: Annotated[
        bool,
        typer.Option(
            ...,
            "--split-ahb",
            "-a",
            help="If set, the AHB will be split into multiple files, one for each Anwendungsfall. "
            "The files will be named `<Prüfidentifikator>.json` in a directory named after the AHB file's "
            "name (without the extension). It will contain an additional `meta.json` file containing the fields of "
            "`Anwendungshandbuch` except for `anwendungsfaelle`.",
        ),
    ] = False,
) -> None:
    """
    Converts the xml file(s) from `xml_in_path` to a json file next to the `*.xml`.
    If `xml_in_path` is a directory, it will search for all XML files in the directory and its subdirectories.

    All xml files must follow the naming convention `/^(?P<FORMAT>[A-Z]+)_(AHB|MIG)_((Gas|Strom)_)?.*\\.xml$/`
    """
    if xml_path.is_dir():
        xml2json_dir_mode(xml_path, sanitize=sanitize, compressed=compressed, split_ahb=split_ahb)
    else:
        xml2json_file_mode(xml_path, sanitize=sanitize, compressed=compressed, split_ahb=split_ahb)



"""
Contains some utility functions that are used in the project.
"""

import re
from typing import Optional

from fundamend.models.anwendungshandbuch import Kommunikationsrichtung


def lstrip(prefix: str, text: str) -> str:
    """Strip the given prefix from the given text. If the text does not start with the prefix, return the text as is.

    Args:
        prefix: The prefix to strip.
        text: The text to strip the prefix from.

    Returns:
        The text with the prefix stripped.
    """
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def rstrip(text: str, suffix: str) -> str:
    """Strip the given suffix from the given text. If the text does not end with the suffix, return the text as is.

    Args:
        text: The text to strip the suffix from.
        suffix: The suffix to strip.

    Returns:
        The text with the suffix stripped.
    """
    if text.endswith(suffix):
        return text[: -len(suffix)]
    return text


def strip(prefix: str, text: str, suffix: str) -> str:
    """Strip the given prefix and suffix from the given text. If the text does not start with the prefix or does not
    end with the suffix, return the text as is.

    Args:
        prefix: The prefix to strip.
        text: The text to strip the prefix from.
        suffix: The suffix to strip.

    Returns:
        The text with the prefix and suffix stripped.
    """
    return lstrip(prefix, rstrip(text, suffix))


_UNIFIED_SEPARATOR = "/"  # how multiple Marktrollen shall be split in the kommunikation_von attribute
_ALTERNATIVE_SEPARATORS = [","]  # other separators that are used in the wild

_an_at_word_boundary = re.compile(r"\ban\b")


def _add_whitespace_before_an(original: str) -> str:
    """adds whitespace before 'an' if it is not already there"""
    return _an_at_word_boundary.sub(" an", original)


def _parse_kommunikation_von_line(kommunikation_von_line: str) -> list[Kommunikationsrichtung]:
    """
    parses a single line of kommunikation_von into a list of Kommunikationsrichtung objects
    this is necessary because some AHBs have multiple lines in the kommunikation_von attribute which must not be mixed
    """
    if not kommunikation_von_line or not kommunikation_von_line.strip():
        return []
    result: list[Kommunikationsrichtung] = []
    parts = _add_whitespace_before_an(kommunikation_von_line).split(" an ")
    if len(parts) != 2:
        # maybe this line looks different, more like 'NB an LF, MSB an NB (Gas)'
        # then we have to split at the comma first and treat each part like it was a single line. wtf
        if "," in kommunikation_von_line:
            for subpart in kommunikation_von_line.split(","):
                result += _parse_kommunikation_von_line(subpart.strip())
            return result
        raise ValueError(f"Invalid kommunikation_von string: '{kommunikation_von_line}'. Expected format: 'X an Y[/Z]'")
    sender_str = parts[0]
    receiver_str = parts[1]
    for alternative_separator in _ALTERNATIVE_SEPARATORS:
        if alternative_separator in receiver_str:
            receiver_str = receiver_str.replace(alternative_separator, _UNIFIED_SEPARATOR)
        if alternative_separator in sender_str:
            sender_str = sender_str.replace(alternative_separator, _UNIFIED_SEPARATOR)
    senders = [x.strip() for x in sender_str.split(_UNIFIED_SEPARATOR)]
    receivers = [x.strip() for x in receiver_str.split(_UNIFIED_SEPARATOR)]
    for sender in senders:
        for receiver in receivers:
            result.append(Kommunikationsrichtung(sender=sender, empfaenger=receiver))
    return result


def parse_kommunikation_von(kommunikation_von: Optional[str]) -> list[Kommunikationsrichtung] | None:
    """Splits the kommunikation_von string into something strongly typed

    Args:
        kommunikation_von: The kommunikation_von string to split, e.g. 'NB an LF/MSB'.

    Returns:
        Properly typed list of Kommunikationsrichtung objects:
        [Kommunikationsrichtung(sender='NB', empfaenger='LF'),
        Kommunikationsrichtung(sender='NB', empfaenger='MSB')]
        or none in case there are no information given (directly).
    """
    if kommunikation_von == "Beteiligte aus Ursprungs-nachricht":
        return None
    result: list[Kommunikationsrichtung] = []
    for line in (kommunikation_von or "").splitlines():
        line = line.strip()
        if line:
            result += _parse_kommunikation_von_line(line)
    return result


__all__ = ["lstrip", "rstrip", "strip", "parse_kommunikation_von"]



"""classes that represent MIGs"""

from datetime import date
from enum import StrEnum

from efoli import EdifactFormat

from .base import FundamendBaseModel

# I didn't invent the data model ;)
# pylint:disable=too-many-instance-attributes


class MigStatus(StrEnum):
    """
    status of a MIG element
    """

    M = "M"
    C = "C"
    R = "R"
    N = "N"
    D = "D"
    O = "O"


class Code(FundamendBaseModel):
    """
    A single code element inside a MIG data element, indicated by the `<Code>` tag.
    """

    # Example:
    # <Code Name="Netznutzungszeiten-Nachricht" Description="">UTILTS</Code>
    name: str  # e.g. 'Netznutzungszeiten-Nachricht'
    description: str | None = None  # e.g. ''
    value: str | None  # e.g. 'UTILTS'


class DataElement(FundamendBaseModel):
    """
    A single data element inside a MIG Segment.
    This models both the 'Datenelement' and the 'Gruppendatenelement', indicated by the `<D_xxxx` tag.
    Are able to contain a single or multiple Code elements.
    """

    # pylint:disable=line-too-long
    # Example:
    # <D_0065 Name="Nachrichtentyp-Kennung" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..6" Format_Specification="an..6">
    #   <Code Name="Netznutzungszeiten-Nachricht" Description="">UTILTS</Code>
    # </D_0065>
    id: str  # e.g. 'D_0065'
    name: str  # e.g. 'Nachrichtentyp-Kennung'
    description: str | None = None  # e.g. ''
    status_std: MigStatus  # e.g. M
    status_specification: MigStatus  # e.g. M
    format_std: str  #: e.g. 'an..6'
    format_specification: str  #: e.g. 'an..6'
    codes: tuple[Code, ...]


class DataElementGroup(FundamendBaseModel):
    """
    A group of data elements, German 'Datenelementgruppe', indicated by the `<C_xxxx>` tag.
    Are able to contain a single or multiple data elements.
    """

    # "Die Datenelementgruppe C0829 enthält mehrere Gruppendatenelemente. Diese Datenelementgruppe enthält das
    # Gruppendatenelement DE3039, hier wird die MP-ID angegeben, sowie das DE3055, welches die code-vergebende Stelle
    # definiert. Diese Datenelementgruppe enthält des Weiteren das Gruppendatenelement DE113110, welches nur in der MIG
    # und nicht im AHB aufgeführt wird, um den Aufbau der EDIFACT Nachricht korrekt umsetzen zu können."
    # Quelle: Allgemeine Festlegungen Kapitel 6.1, Seite 52

    # Example:
    # pylint:disable=line-too-long
    # <C_C082 Name="Identifikation des Beteiligten" Description="" Status_Std="C" Status_Specification="R">
    #   <D_3039 Name="MP-ID" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..35" Format_Specification="an..35"/>
    #   <D_1131 Name="Codeliste, Code" Description="" Status_Std="C" Status_Specification="N" Format_Std="an..17" Format_Specification="an..17"/>
    #   <D_3055 Name="Verantwortliche Stelle für die Codepflege, Code" Description="" Status_Std="C" Status_Specification="R" Format_Std="an..3" Format_Specification="an..3">
    #     <Code Name="GS1" Description="">9</Code>
    #     <Code Name="DE, BDEW (Bundesverband der Energie- und Wasserwirtschaft e.V.)" Description="">293</Code>
    #   </D_3055>
    # </C_C082>

    id: str  # e.g. 'C_C082'
    name: str  # e.g. 'Identifikation des Beteiligten'
    description: str | None = None  # e.g. ''
    status_std: MigStatus  # e.g. C
    status_specification: MigStatus  # e.g. R
    data_elements: tuple[DataElement, ...]


class Segment(FundamendBaseModel):
    """
    A segment inside a MIG, indicated by the `<S_xxxx>` tag. A segment contains data elements and data element groups.
    """

    # Example:
    # pylint:disable=line-too-long
    # <S_NAD Name="MP-ID Absender" Description="DE3039: Zur Identifikation der Marktpartner wird die MP-ID angegeben." Counter="0100" Level="1" Number="00004" MaxRep_Std="1" MaxRep_Specification="1" Status_Std="M" Status_Specification="M" Example="NAD+MS+9900259000002::293&apos;">
    #     <D_3035 Name="Beteiligter, Qualifier" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..3" Format_Specification="an..3">
    #        <Code Name="Dokumenten-/Nachrichtenaussteller bzw. -absender" Description="">MS</Code>
    #     </D_3035>
    #     <C_C082 Name="Identifikation des Beteiligten" Description="" Status_Std="C" Status_Specification="R">
    #       <D_3039 Name="MP-ID" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..35" Format_Specification="an..35"/>
    #       <D_1131 Name="Codeliste, Code" Description="" Status_Std="C" Status_Specification="N" Format_Std="an..17" Format_Specification="an..17"/>
    #       <D_3055 Name="Verantwortliche Stelle für die Codepflege, Code" Description="" Status_Std="C" Status_Specification="R" Format_Std="an..3" Format_Specification="an..3">
    #         <Code Name="GS1" Description="">9</Code>
    #         <Code Name="DE, BDEW (Bundesverband der Energie- und Wasserwirtschaft e.V.)" Description="">293</Code>
    #      </D_3055>
    #   </C_C082>
    # </S_NAD>
    id: str  #: e.g. 'NAD'
    name: str  #: e.g. 'MP-ID Absender'
    description: str | None = None  # e.g. 'DE3039: Zur Identifikation der Marktpartner wird die MP-ID angegeben.'
    counter: str  #: e.g. '0100'
    level: int  #: e.g. 1
    number: str  #: e.g. '00004'
    max_rep_std: int  #: e.g. 1
    max_rep_specification: int  #: e.g. 1
    status_std: MigStatus  # e.g. M
    status_specification: MigStatus  # e.g. M
    example: str | None  #: e.g. "NAD+MS+9900259000002::293'"
    data_elements: tuple[DataElement | DataElementGroup, ...]
    is_on_uebertragungsdatei_level: bool = False
    """
    true only for those segments that are on the 'UEBERTRAGUNGSDATEI' level (meaning, outside the <M_FORMAT> tag).
    Similar to the corresponding AHB tag, we decided against replicating this XML level in the fundamend data model.
    This is the case for the UNA and UNZ segments in UTILMD, REMADV or INVOIC.
    But other formats like UTILTS or ORDERS have no such level at all.
    Basically this flag describes on which level you may aggregate single "Geschäftsvorfälle"/transactions in a single
    Übertragungsdatei.
    You may, for example, send multiple UTILMDs oR MSCONS messages batched/aggregated in one file, but not multiple
    ORDERs.
    """


class SegmentGroup(FundamendBaseModel):
    """
    A 'Segmentgruppe' inside a MIG, indicated by the `<G_xxx>` tag. A segment contains segments and segments groups.
    """

    # pylint:disable=line-too-long
    # Example:
    # <G_SG2 Name="MP-ID Empfänger" Counter="0090" Level="1" MaxRep_Std="99" MaxRep_Specification="1" Status_Std="C" Status_Specification="R">
    #   <S_NAD Name="MP-ID Empfänger" Description="DE3039: Zur Identifikation der Marktpartner wird die MP-ID angegeben." Counter="0100" Level="1" Number="00007" MaxRep_Std="1" MaxRep_Specification="1" Status_Std="M" Status_Specification="M" Example="NAD+MR+9900259000002::293&apos;">
    #     <D_3035 Name="Beteiligter, Qualifier" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..3" Format_Specification="an..3">
    #        <Code Name="Nachrichtenempfänger" Description="">MR</Code>
    #     </D_3035>
    #     <C_C082 Name="Identifikation des Beteiligten" Description="" Status_Std="C" Status_Specification="R">
    #       <D_3039 Name="MP-ID" Description="" Status_Std="M" Status_Specification="M" Format_Std="an..35" Format_Specification="an..35"/>
    #       <D_1131 Name="Codeliste, Code" Description="" Status_Std="C" Status_Specification="N" Format_Std="an..17" Format_Specification="an..17"/>
    #       <D_3055 Name="Verantwortliche Stelle für die Codepflege, Code" Description="" Status_Std="C" Status_Specification="R" Format_Std="an..3" Format_Specification="an..3">
    #         <Code Name="GS1" Description="">9</Code>
    #         <Code Name="DE, BDEW (Bundesverband der Energie- und Wasserwirtschaft e.V.)" Description="">293</Code>
    #       </D_3055>
    #     </C_C082>
    #   </S_NAD>
    # </G_SG2>
    id: str  # e.g. 'SG2'
    name: str  # e.g. 'MP-ID Empfänger'
    counter: str  # e.g. '0090'
    level: int  # e.g. 1
    max_rep_std: int  # e.g. 99
    max_rep_specification: int  # e.g. 1
    status_std: MigStatus
    status_specification: MigStatus
    elements: tuple["Segment | SegmentGroup", ...]


class MessageImplementationGuide(FundamendBaseModel):
    """
    message implementation guide (MIG)
    """

    veroeffentlichungsdatum: date
    """publishing date"""

    autor: str
    """author, most likely 'BDEW'"""

    versionsnummer: str
    """e.g. '1.1c'"""

    format: EdifactFormat  #: e.g. 'UTILTS'

    elements: tuple[Segment | SegmentGroup, ...]



"""
This module contains the SQLModel class for the AHBesser view and a function to create it.
If you never heard about ahbesser, you can safely ignore this module.
https://github.com/Hochfrequenz/ahbesser
"""

import logging
from pathlib import Path
from uuid import UUID

from efoli import EdifactFormat, EdifactFormatVersion
from sqlalchemy.sql.functions import func
from sqlmodel import Field, Session, SQLModel, select

from fundamend.sqlmodels.internals import _execute_bare_sql

_logger = logging.getLogger(__name__)


def create_ahbtabellen_view(session: Session) -> None:
    """
    Create a view for the AHB-Tabellen application: https://github.com/Hochfrequenz/ahbesser
    This assumes that create_db_and_populate_with_ahb_view has already been called.
    If you don't know what ahbesser is, you can safely ignore this function.
    """
    _execute_bare_sql(session=session, path_to_sql_commands=Path(__file__).parent / "create_ahbtabellen_view.sql")
    number_of_rows = session.scalar(
        select(func.count(AhbTabellenLine.id))  # type:ignore[arg-type] # pylint:disable=not-callable
    )
    _logger.info(
        "There are %d rows in the AHBTabellen view %s",
        number_of_rows,
        AhbTabellenLine.__tablename__,
    )


class AhbTabellenLine(SQLModel, table=True):
    """
    Model that represents thew view used by ahbesser. It's created by executing 'create_ahbtabellen_view(session)'
    """

    __tablename__ = "v_ahbtabellen"
    id: UUID = Field(primary_key=True)
    format_version: EdifactFormatVersion = Field()
    format: EdifactFormat = Field()
    pruefidentifikator: str = Field()
    path: str = Field()
    id_path: str = Field()
    direction: str = Field()
    description: str = Field()
    segmentgroup_key: str | None = Field()
    segment_code: str | None = Field()
    data_element: str | None = Field()
    qualifier: str | None = Field()
    is_on_uebertragungsdatei_level: bool | None = Field()  #: !=null as soon as at least one segment is part of the path
    line_ahb_status: str | None = Field()
    line_name: str | None = Field()
    line_type: str | None = Field()  # 'code' or 'dataelement' or 'segment' or 'segmentgroup' or 'dataelementgroup'
    bedingung: str | None = Field()
    bedingungsfehler: str | None = Field()
    sort_path: str = Field()


__all__ = ["create_ahbtabellen_view", "AhbTabellenLine"]



"""
helper module to create a table with a "Bedingung" column like the one in the PDF/docx AHBs
"""

import asyncio
import logging
import uuid
from contextvars import ContextVar
from typing import Optional

from efoli import EdifactFormat, EdifactFormatVersion

from fundamend.sqlmodels import AhbHierarchyMaterialized, Bedingung
from fundamend.sqlmodels.anwendungshandbuch import Paket, UbBedingung

try:
    from sqlalchemy.sql.functions import func
    from sqlmodel import Field, Session, SQLModel, UniqueConstraint, col, select

except ImportError as import_error:
    import_error.msg += "; Did you install fundamend[sqlmodels] or did you try to import from fundamend.models instead?"
    # sqlmodel is only an optional dependency when fundamend is used to fill a database
    raise


try:
    import inject
    from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluatableDataProvider
    from ahbicht.content_evaluation.evaluator_factory import create_content_evaluation_result_based_evaluators
    from ahbicht.content_evaluation.expression_check import is_valid_expression
    from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider, TokenLogicProvider
    from ahbicht.expressions.condition_expression_parser import extract_categorized_keys
    from ahbicht.models.content_evaluation_result import ContentEvaluationResult, ContentEvaluationResultSchema
    from lark.exceptions import VisitError
except ImportError as import_error:
    import_error.msg += "; Did you install fundamend[sqlmodels,ahbicht]?"
    # sqlmodel and ahbicht are only optional dependencies when fundamend is used to fill a database
    raise

_logger = logging.getLogger(__name__)

_content_evaluation_result: ContextVar[Optional[ContentEvaluationResult]] = ContextVar(
    "_content_evaluation_result", default=None
)


def _get_evaluatable_data() -> EvaluatableData[ContentEvaluationResult]:
    """
    returns the _content_evaluation_result context var value wrapped in a EvaluatableData container.
    This is the kind of data that the ContentEvaluationResultBased RC/FC Evaluators, HintsProvider and Package Resolver
    require.
    :return:
    """
    cer = _content_evaluation_result.get()
    return EvaluatableData(
        body=ContentEvaluationResultSchema().dump(cer),
        edifact_format=EdifactFormat.UTILMD,  # not important, something has to be here
        edifact_format_version=EdifactFormatVersion.FV2504,  # not important, something has to be here
    )


def _setup_weird_ahbicht_dependency_injection() -> None:
    def configure(binder: inject.Binder) -> None:
        binder.bind(
            TokenLogicProvider,
            SingletonTokenLogicProvider(
                [*create_content_evaluation_result_based_evaluators(EdifactFormat.UTILMD, EdifactFormatVersion.FV2504)]
            ),
        )
        binder.bind_to_provider(EvaluatableDataProvider, _get_evaluatable_data)

    inject.configure_once(configure)


def _generate_node_texts(session: Session, expression: str, ahb_pk: uuid.UUID) -> str:
    categorized_key_extract = asyncio.run(extract_categorized_keys(expression))
    bedingung_keys = (
        categorized_key_extract.format_constraint_keys
        + categorized_key_extract.requirement_constraint_keys
        + categorized_key_extract.hint_keys
    )
    paket_keys = categorized_key_extract.package_keys
    ubbedingung_keys = categorized_key_extract.time_condition_keys
    # probably, we'd be faster if we just loaded all pakete and all bedingungen once instead of selecting over and over
    # again for each expression
    bedingungen = {
        x.nummer: x.text
        for x in session.exec(
            select(Bedingung).where(
                col(Bedingung.nummer).in_(bedingung_keys),  # pylint:disable=no-member
                Bedingung.anwendungshandbuch_primary_key == ahb_pk,
            )
        ).all()
    }
    pakete = {
        x.nummer: x.text
        for x in session.exec(
            select(Paket).where(
                col(Paket.nummer).in_(paket_keys),
                Paket.anwendungshandbuch_primary_key == ahb_pk,  # pylint:disable=no-member
            )
        ).all()
    }
    ubbedingungen = {
        x.nummer: x.text
        for x in session.exec(
            select(UbBedingung).where(
                col(UbBedingung.nummer).in_(ubbedingung_keys),  # pylint:disable=no-member
                UbBedingung.anwendungshandbuch_primary_key == ahb_pk,
            )
        ).all()
    }
    joined_dict = {**bedingungen, **pakete, **ubbedingungen}
    node_texts = "\n".join([f"[{key}] {value}" for key, value in joined_dict.items()])
    return node_texts


def _get_validity_node_texts_and_error_message_cpu_intensive(
    expression: str, session: Session, anwendungshandbuch_pk: uuid.UUID
) -> tuple[bool, str, str | None]:
    try:
        is_valid, error_message = asyncio.run(is_valid_expression(expression, _content_evaluation_result.set))
        if is_valid:  # we might actually get a meaningful node_texts even for invalid expressions, but I don't like it
            node_texts = _generate_node_texts(session, expression, anwendungshandbuch_pk)
        else:
            node_texts = ""
    except NotImplementedError:  # ahbicht fault/missing feature -> act like it's valid
        node_texts = _generate_node_texts(session, expression, anwendungshandbuch_pk)
        error_message = None
    return is_valid, node_texts, error_message


def _get_validity_node_texts_and_error_message_fast(
    expression: str, session: Session, anwendungshandbuch_pk: uuid.UUID
) -> tuple[bool, str, str | None]:
    try:
        node_texts = _generate_node_texts(session, expression, anwendungshandbuch_pk)
    except SyntaxError as syntax_error:
        _logger.info("The expression '%s' could not be parsed: %s", expression, syntax_error)
        return (
            False,
            "",
            str(syntax_error),
        )  # I decided against returning the error message, although it's tempting - but still bad practice
    except VisitError as visit_error:
        _logger.info("The expression '%s' could not be parsed: %s", expression, visit_error)
        return False, "", str(visit_error)
    return True, node_texts, None


def create_and_fill_ahb_expression_table(session: Session, use_cpu_intensive_validity_check: bool = False) -> None:
    """
    creates and fills the ahb_expressions table. It uses the ahb_hierarchy_materialized table to extract all expressions
    and parses each expression with ahbicht. The latter has to be done in Python.
    If the CPU intensive validity check is enabled, not only expression alone is checked but also all its possible
    outcomes. This leads to only few additional expressions marked as invalid but is very slow.
    """
    rows: list[tuple[EdifactFormatVersion | None, str, str | None, uuid.UUID]] = []
    _setup_weird_ahbicht_dependency_injection()
    for ahb_status_col in [
        AhbHierarchyMaterialized.segmentgroup_ahb_status,
        AhbHierarchyMaterialized.segment_ahb_status,
        AhbHierarchyMaterialized.dataelement_ahb_status,
        AhbHierarchyMaterialized.code_ahb_status,
    ]:
        stmt = select(
            AhbHierarchyMaterialized.edifact_format_version,
            AhbHierarchyMaterialized.format,
            ahb_status_col,
            AhbHierarchyMaterialized.anwendungshandbuch_primary_key,
        )
        rows.extend(session.exec(stmt))  # type:ignore[arg-type]
    non_empty_rows: list[tuple[EdifactFormatVersion, str, str, uuid.UUID]] = [
        r for r in rows if r[2] is not None and r[0] is not None and r[2].strip()  # type:ignore[misc]
    ]
    if not any(rows):
        raise ValueError(
            "No rows found in ahb_hierarchy_materialized table; Run `create_db_and_populate_with_ahb_view` before."
        )
    non_empty_rows.sort(key=lambda x: (x[0], x[1], x[2]))
    seen: set[tuple[str, str, str]] = set()
    unique_rows = [
        row
        for row in non_empty_rows
        if (key := (row[0], row[1], row[2].strip())) not in seen
        and not seen.add(key)  # type:ignore[ func-returns-value]
    ]
    for row in unique_rows:  # there are ~3600 unique rows for FV2410+FV2504 as of 2025-04-15
        expression = row[2].strip()
        if use_cpu_intensive_validity_check:
            # as of 2025-04-15 I have no clue how long this actually takes for all expressions
            _, node_texts, error_message = _get_validity_node_texts_and_error_message_cpu_intensive(
                expression, session, row[3]
            )
        else:
            _, node_texts, error_message = _get_validity_node_texts_and_error_message_fast(expression, session, row[3])
        ahb_expression_row = AhbExpression(
            edifact_format_version=row[0],
            format=row[1],
            expression=expression,
            node_texts=node_texts,
            anwendungshandbuch_primary_key=row[3],
            ahbicht_error_message=error_message,
        )
        session.add(ahb_expression_row)
        _logger.debug(
            "Added row (%s, %s, %s) to the ahb_expressions_table",
            ahb_expression_row.edifact_format_version,
            ahb_expression_row.format,
            ahb_expression_row.expression,
        )
    number_of_inserted_rows = session.scalar(
        select(func.count(AhbExpression.id))  # type:ignore[arg-type]# pylint:disable=not-callable
    )
    _logger.info(
        "Inserted %d rows into the table %s",
        number_of_inserted_rows,
        AhbExpression.__tablename__,
    )
    session.commit()


class AhbExpression(SQLModel, table=True):
    """
    A table that contains all expressions that are used in any AHB, each with prüfidentifikator and format_version.
    It's created by UNIONing all 'ahb_status' columns from all relevant tables.
    Additionally, this table has a column that resolves the expression to a human-readable text.
    """

    __tablename__ = "ahb_expressions"
    __table_args__ = (
        UniqueConstraint(
            "edifact_format_version",
            "format",
            "expression",
            name="idx_ahb_expressions_metadata_expression",
        ),
    )
    id: uuid.UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    edifact_format_version: EdifactFormatVersion = Field(index=True)
    format: EdifactFormat = Field(index=True)  # the edifact format, e.g. 'UTILMD'
    # expressions and conditions are always interpreted on a per-format basis (no pruefidentifikator required)
    expression: str = Field(index=True)  #: e.g 'Muss [1] U [2]'
    node_texts: str = Field()
    """
    this contains the typical "[1] Foo Text\n[2] Bar Text" which explains the meaning of the nodes from inside the
    respective Expression (e.g. for expression "Muss [1] U [2]")
    """
    ahbicht_error_message: str | None = Field(default=None)
    anwendungshandbuch_primary_key: uuid.UUID = Field()



"""shared logic between MIG and AHB reader to distinguish elements from the XML"""

import xml.etree.ElementTree as ET


def _is_segment_group(element: ET.Element) -> bool:
    """
    returns True if the given element is a segment group
    """
    return element.tag.startswith("G_SG")


def _is_segment(element: ET.Element) -> bool:
    """
    returns True if the given element is a segment
    """
    return element.tag.startswith("S_")


def _is_data_element_group(element: ET.Element) -> bool:
    """
    returns True if the given element is a data element group
    """
    return element.tag.startswith("C_")


def _is_data_element(element: ET.Element) -> bool:
    """
    returns True if the given element is a data element
    """
    return element.tag.startswith("D_")


def _is_code(element: ET.Element) -> bool:
    """
    returns True if the given element is a code
    """
    return element.tag == "Code"


def _is_format(element: ET.Element) -> bool:
    """returns true if element is wrapper around an anwendungsfall"""
    return element.tag.startswith("M_")


def _is_anwendungsfall(element: ET.Element) -> bool:
    """returns true iff the element is an AHB anwendungsfall"""
    return element.tag == "AWF"


def _is_uebertragungsdatei(element: ET.Element) -> bool:
    """returns true iff the element is an Uebertragsdatei element (usually below the format level)"""
    return element.tag == "Uebertragungsdatei"



"""contains model classes representing MIGs and AHBs"""

from .anwendungshandbuch import Anwendungshandbuch
from .messageimplementationguide import MessageImplementationGuide

__all__ = ["MessageImplementationGuide", "Anwendungshandbuch"]



"""
This module provides functions to sanitize the AHB (Anwendungshandbuch) by adding
unused MIG (Message Implementation Guide) elements to the AHB.
They will be marked with an ahb_status of ``X [2499]`` which should always evaluate to false aka "must not"
in the AHB expression language.

Note that to save a big chunk of irrelevant data, elements which are marked with ``X [2499]`` will never contain
the sub elements from the MIG. Except for the leading segment of a segment group - this helps to build search indices
for segment groups by using the leading segment's number.
"""

from types import MethodType
from typing import Any, Iterator, overload

from pydantic import BaseModel

from fundamend.models import anwendungshandbuch as ahb
from fundamend.models import messageimplementationguide as mig


def _disabled_hash(_: Any) -> int:  # pragma: no cover
    raise ValueError("Hash function is disabled for this model as some attribute was overridden by object.__setattr__.")


def _set(model: BaseModel, field_name: str, field_value: Any) -> None:
    object.__setattr__(model, field_name, field_value)
    model.__hash__ = MethodType(_disabled_hash, model)  # type: ignore[method-assign]
    # This hash function override is just for our safety to prevent obscure errors when trying to use
    # the model as a hashable object.


@overload
def leading_segment(segment_group: ahb.SegmentGroup) -> ahb.Segment: ...


@overload
def leading_segment(segment_group: mig.SegmentGroup) -> mig.Segment: ...


def leading_segment(
    segment_group: ahb.SegmentGroup | mig.SegmentGroup,
) -> ahb.Segment | mig.Segment:
    """
    Return the first segment of the segment group.

    :param segment_group: The segment group.
    :returns: The first segment of the segment group.
    :raises TypeError: If the segment group doesn't start with a segment. This shouldn't happen if our assumptions are
        correct.
    """

    if not isinstance(segment_group.elements[0], (ahb.Segment, mig.Segment)):  # pragma: no cover
        raise TypeError(
            f"Expected segment group to start with a segment, got {type(segment_group.elements[0])} instead."
        )
    return segment_group.elements[0]


def get_number(obj: mig.SegmentGroup | mig.Segment | ahb.SegmentGroup | ahb.Segment) -> str:
    """
    Get the number of the given segment or segment group.

    In case of a segment group, the number of the leading segment is returned.

    :param obj: The object to get the number from.
    :returns: The number of the segment or the segment groups leading segment.
    :raises ValueError: If the object type is unexpected.
    """
    match obj:
        case mig.SegmentGroup() | ahb.SegmentGroup():
            return leading_segment(obj).number
        case mig.Segment() | ahb.Segment():
            return obj.number
        case _:  # pragma: no cover
            raise ValueError(f"Unexpected object type: {type(obj)}")


def match_ahb_data_element_with_mig_data_element(
    mig_data_element: mig.DataElement, ahb_data_element: ahb.DataElement
) -> bool:
    """
    Check if the given AHB or MIG data element matches the MIG data element.

    :param ahb_data_element: The AHB or MIG data element to check.
    :param mig_data_element: The MIG data element to check against.
    :return: ``True`` if the AHB element matches the MIG data element, ``False`` otherwise.
    """
    return ahb_data_element.id == mig_data_element.id


def match_ahb_data_element_group_with_mig_data_element_group(
    mig_data_element_group: mig.DataElementGroup, ahb_data_element_group: ahb.DataElementGroup
) -> bool:
    """
    Check if the given AHB data element group matches the MIG data element group.

    :param ahb_data_element_group: The AHB data element group to check.
    :param mig_data_element_group: The MIG data element group to check against.
    :return: ``True`` if the AHB data element group matches the MIG data element group, ``False`` otherwise.
    """
    _has_same_id = ahb_data_element_group.id == mig_data_element_group.id
    if not _has_same_id:
        return False
    _all_data_elements_in_mig_data_element_group = all(
        any(
            match_ahb_data_element_with_mig_data_element(mig_data_element, ahb_data_element)
            for mig_data_element in mig_data_element_group.data_elements
        )
        for ahb_data_element in ahb_data_element_group.data_elements
    )
    return _all_data_elements_in_mig_data_element_group


@overload
def parallel_iter_segment_or_data_element_group(
    mig_segment_or_group: mig.Segment, ahb_segment_or_group: ahb.Segment
) -> Iterator[
    tuple[mig.DataElementGroup, ahb.DataElementGroup | None] | tuple[mig.DataElement, ahb.DataElement | None]
]: ...


@overload
def parallel_iter_segment_or_data_element_group(
    mig_segment_or_group: mig.DataElementGroup, ahb_segment_or_group: ahb.DataElementGroup
) -> Iterator[tuple[mig.DataElement, ahb.DataElement | None]]: ...


@overload
def parallel_iter_segment_or_data_element_group(
    mig_segment_or_group: mig.Segment | mig.DataElementGroup, ahb_segment_or_group: ahb.Segment | ahb.DataElementGroup
) -> Iterator[
    tuple[mig.DataElementGroup, ahb.DataElementGroup | None] | tuple[mig.DataElement, ahb.DataElement | None]
]: ...


def parallel_iter_segment_or_data_element_group(
    mig_segment_or_group: mig.Segment | mig.DataElementGroup, ahb_segment_or_group: ahb.Segment | ahb.DataElementGroup
) -> Iterator[
    tuple[mig.DataElementGroup, ahb.DataElementGroup | None] | tuple[mig.DataElement, ahb.DataElement | None]
]:
    """
    If an element is unused in AHB, i.e. if it has MigStatus.N or if it is part of a degenerated element set which
    would result in an IndexError, the MIG element will be yielded with ``None`` as the AHB counterpart.

    :param mig_segment_or_group: The MIG segment or data element group to iterate over.
    :param ahb_segment_or_group: The AHB segment or data element group to iterate over.
    :returns: An iterator yielding pairs of MIG and AHB elements.
    """
    assert len(mig_segment_or_group.data_elements) > 0
    # assert len(mig_segment_or_group.data_elements) > 0 and len(ahb_segment_or_group.data_elements) > 0
    # This assertion doesn't have to always hold. Counter example:
    # ORDERS 17112 G_SG29 LIN'00072'
    mig_elements = iter(mig_segment_or_group.data_elements)
    ahb_elements = iter(ahb_segment_or_group.data_elements)

    cur_mig_element = next(mig_elements)  # pylint: disable=stop-iteration-return
    # This statement actually cannot raise a StopIteration, because we already checked that the MIG segment
    # or data element group has at least one element.
    cur_ahb_element = next(ahb_elements, None)

    while True:
        if isinstance(cur_mig_element, mig.DataElement):
            ahb_match = isinstance(cur_ahb_element, ahb.DataElement) and match_ahb_data_element_with_mig_data_element(
                cur_mig_element, cur_ahb_element
            )
        else:
            assert isinstance(cur_mig_element, mig.DataElementGroup)
            ahb_match = isinstance(
                cur_ahb_element, ahb.DataElementGroup
            ) and match_ahb_data_element_group_with_mig_data_element_group(cur_mig_element, cur_ahb_element)
        if ahb_match:
            yield cur_mig_element, cur_ahb_element  # type: ignore[misc]
            # No assertion needed here, because we already checked the instance types through the `ahb_match` variable.
            cur_ahb_element = next(ahb_elements, None)
        else:
            yield cur_mig_element, None  # type: ignore[misc]
            # mypy is not smart enough to understand the union construction here
        try:
            cur_mig_element = next(mig_elements)
        except StopIteration:
            break
    try:  # pragma: no cover
        next(ahb_elements)
        raise ValueError(
            f"AHB segment or group {ahb_segment_or_group.id} has more data elements than "
            f"MIG segment or group {mig_segment_or_group.id}."
        )
    except StopIteration:
        pass


def parallel_iter_segment_group_or_root(
    mig_segment_group_or_root: mig.MessageImplementationGuide | mig.SegmentGroup,
    ahb_segment_group_or_root: ahb.Anwendungsfall | ahb.SegmentGroup,
) -> Iterator[tuple[mig.SegmentGroup, ahb.SegmentGroup | None] | tuple[mig.Segment, ahb.Segment | None]]:
    """
    Iterates over the elements of the MIG and AHB segment group or root, yielding pairs of MIG and AHB elements.
    If a MIG element is unused in the AHB, it will be yielded with ``None`` as the AHB counterpart.

    :param mig_segment_group_or_root: The MIG segment group or root to iterate over.
    :param ahb_segment_group_or_root: The AHB segment group or root to iterate over.
    :returns: An iterator yielding pairs of MIG and AHB elements.
    """
    assert len(mig_segment_group_or_root.elements) > 0 and len(ahb_segment_group_or_root.elements) > 0
    mig_elements = iter(mig_segment_group_or_root.elements)
    ahb_elements = iter(ahb_segment_group_or_root.elements)

    cur_mig_element = next(mig_elements)  # pylint: disable=stop-iteration-return
    cur_ahb_element = next(ahb_elements)  # pylint: disable=stop-iteration-return
    # This statement actually cannot raise a StopIteration, because we already checked that the MIG segment
    # or data element group has at least one element.

    while True:
        mig_number = get_number(cur_mig_element)
        ahb_number = cur_ahb_element and get_number(cur_ahb_element)
        if mig_number == ahb_number:
            assert (
                isinstance(cur_mig_element, mig.Segment)
                and isinstance(cur_ahb_element, ahb.Segment)
                or isinstance(cur_mig_element, mig.SegmentGroup)
                and isinstance(cur_ahb_element, ahb.SegmentGroup)
            )
            yield cur_mig_element, cur_ahb_element  # type: ignore[misc]
            # mypy is not smart enough to understand the union construction here
            cur_ahb_element = next(ahb_elements, None)  # type: ignore[assignment]
        else:
            yield cur_mig_element, None  # type: ignore[misc]
            # mypy is not smart enough to understand the union construction here
        try:
            cur_mig_element = next(mig_elements)
        except StopIteration:
            break
    try:  # pragma: no cover
        next(ahb_elements)
        ahb_element_id = (
            f"{ahb_segment_group_or_root.id}'{get_number(ahb_segment_group_or_root)}'"
            if isinstance(ahb_segment_group_or_root, ahb.SegmentGroup)
            else "root"
        )
        mig_element_id = (
            f"{mig_segment_group_or_root.id}'{get_number(mig_segment_group_or_root)}'"
            if isinstance(mig_segment_group_or_root, mig.SegmentGroup)
            else "root"
        )
        raise ValueError(f"AHB {ahb_element_id} has more elements than MIG {mig_element_id}.")
    except StopIteration:
        pass


def create_ahb_code_from_mig(mig_code: mig.Code) -> ahb.Code:
    """
    Creates an AHB code from a MIG code.

    :param mig_code: The MIG code to convert.
    :return: An AHB code with the same name, description and value.
    """
    return ahb.Code(
        name=mig_code.name,
        description=mig_code.description,
        value=mig_code.value,
        ahb_status="X [2499]",
    )


def create_ahb_data_element_from_mig(mig_data_element: mig.DataElement) -> ahb.DataElement:
    """
    Creates an AHB data element from a MIG data element.
    This is used to add unused MIG data elements to the AHB.

    Sometimes a MIG data element can have a list of codes, but is still unused in the AHB. An example for this is
    ``CONTRL UCI D_0085``.
    In this case, we still don't add the codes to the AHB data element, because the data element will be forbidden
    by the AHB expression anyway.

    :param mig_data_element: The MIG data element to convert.
    :return: An AHB data element with the same ID and name, but with an ahb_status of ``X [2499]`` and no codes.
    """
    return ahb.DataElement(
        id=mig_data_element.id,
        name=mig_data_element.name,
        ahb_status="X [2499]" if len(mig_data_element.codes) == 0 else None,
        codes=tuple(create_ahb_code_from_mig(mig_code) for mig_code in mig_data_element.codes),
    )


def create_ahb_data_element_group_from_mig(mig_data_element_group: mig.DataElementGroup) -> ahb.DataElementGroup:
    """
    Creates an AHB data element group from a MIG data element group.
    This is used to add unused MIG data element groups to the AHB.

    :param mig_data_element_group: The MIG data element group to convert.
    :return: An AHB data element group with the same ID and name. The data elements of the MIG data element group
        are converted to AHB data elements by ``create_ahb_data_element_from_mig``.
    """
    return ahb.DataElementGroup(
        id=mig_data_element_group.id,
        name=mig_data_element_group.name,
        data_elements=tuple(
            create_ahb_data_element_from_mig(mig_data_element)
            for mig_data_element in mig_data_element_group.data_elements
        ),
    )


def create_ahb_segment_from_mig(mig_segment: mig.Segment) -> ahb.Segment:
    """
    Creates an AHB segment from a MIG segment.
    This is used to add unused MIG segments to the AHB.

    :param mig_segment: The MIG segment to convert.
    :return: An AHB segment with the same ID, name and number, but with an ahb_status of ``X [2499]`` and
        no data elements. We can safely omit the data elements here, because an application can detect
        the unused segment by the ``ahb_status`` and stop further traversal.
    """
    return ahb.Segment(
        id=mig_segment.id,
        name=mig_segment.name,
        number=mig_segment.number,
        ahb_status="X [2499]",
        data_elements=tuple(
            (
                create_ahb_data_element_from_mig(mig_element)
                if isinstance(mig_element, mig.DataElement)
                else create_ahb_data_element_group_from_mig(mig_element)
            )
            for mig_element in mig_segment.data_elements
        ),
    )


def create_ahb_segment_group_from_mig(mig_segment_group: mig.SegmentGroup) -> ahb.SegmentGroup:
    """
    Creates an AHB segment group from a MIG segment group.
    This is used to add unused MIG segment groups to the AHB.

    :param mig_segment_group: The MIG segment group to convert.
    :return: An AHB segment group with the same ID and name, but with an ahb_status of ``X [2499]``. To enable
        building search indices for segment groups based on the leading segments number, the first segment of the
        MIG segment group is also converted to an AHB segment and added to the AHB segment group. All other segments
        will be omitted since an application can detect the unused segment group by the ``ahb_status`` and stop further
        traversal.
    """
    return ahb.SegmentGroup(
        id=mig_segment_group.id,
        name=mig_segment_group.name,
        ahb_status="X [2499]",
        elements=tuple(
            (
                create_ahb_segment_from_mig(mig_element)
                if isinstance(mig_element, mig.Segment)
                else create_ahb_segment_group_from_mig(mig_element)
            )
            for mig_element in mig_segment_group.elements
        ),
    )


def add_unused_data_elements_or_groups_to_ahb(
    mig_element: mig.Segment | mig.DataElementGroup, ahb_element: ahb.Segment | ahb.DataElementGroup
) -> None:
    """
    Adds unused MIG data elements or data element groups to the AHB data element group or segment.
    This ensures that all MIG elements are present in the AHB for easy parallel iteration of applications.

    :param mig_element: The MIG segment or data element group to add unused elements from.
    :param ahb_element: The AHB segment or data element group to add unused elements to.
    """
    edited_ahb_elements = list(ahb_element.data_elements)
    index = 0
    for cur_mig_element, cur_ahb_element in parallel_iter_segment_or_data_element_group(mig_element, ahb_element):
        if cur_ahb_element is None:
            # MIG element is unused in AHB, add it to AHB
            if isinstance(cur_mig_element, mig.DataElement):
                edited_ahb_elements.insert(index, create_ahb_data_element_from_mig(cur_mig_element))
            else:
                assert isinstance(cur_mig_element, mig.DataElementGroup)
                edited_ahb_elements.insert(index, create_ahb_data_element_group_from_mig(cur_mig_element))
        elif isinstance(cur_mig_element, mig.DataElementGroup):
            assert isinstance(cur_ahb_element, ahb.DataElementGroup)
            add_unused_data_elements_or_groups_to_ahb(cur_mig_element, cur_ahb_element)
        index += 1
    _set(ahb_element, "data_elements", tuple(edited_ahb_elements))
    assert len(ahb_element.data_elements) == len(mig_element.data_elements)


def add_unused_segment_or_groups_to_ahb(
    mig_root: mig.MessageImplementationGuide | mig.SegmentGroup, ahb_root: ahb.Anwendungsfall | ahb.SegmentGroup
) -> None:
    """
    Adds unused MIG segments or segment groups to the AHB segment group or root element.
    This ensures that all MIG elements are present in the AHB for easy parallel iteration of applications.
    The ``add_unused_data_elements_or_groups_to_ahb`` function is called recursively for segments which are present
    in both the MIG and AHB.

    :param mig_root: The MIG segment group or root to add unused elements from.
    :param ahb_root: The AHB segment group or root to add unused elements to.
    """
    edited_ahb_elements = list(ahb_root.elements)
    index = 0
    for mig_element, ahb_element in parallel_iter_segment_group_or_root(mig_root, ahb_root):
        if ahb_element is None:
            if isinstance(mig_element, mig.Segment):
                edited_ahb_elements.insert(index, create_ahb_segment_from_mig(mig_element))
            else:
                assert isinstance(mig_element, mig.SegmentGroup)
                edited_ahb_elements.insert(index, create_ahb_segment_group_from_mig(mig_element))
        else:
            if isinstance(mig_element, mig.Segment):
                assert isinstance(ahb_element, ahb.Segment)
                add_unused_data_elements_or_groups_to_ahb(mig_element, ahb_element)
            elif isinstance(mig_element, mig.SegmentGroup):
                assert isinstance(ahb_element, ahb.SegmentGroup)
                add_unused_segment_or_groups_to_ahb(mig_element, ahb_element)
        index += 1
    _set(ahb_root, "elements", tuple(edited_ahb_elements))
    assert len(ahb_root.elements) == len(mig_root.elements)


def add_must_not_pattern_to_ahb_conditions(ahb_root: ahb.Anwendungshandbuch) -> None:
    """
    Adds the condition for the must-not pattern to the AHB conditions.
    This condition is used to mark elements which are unused in the AHB and should not be provided in a message.

    :param ahb_root: The AHB root element to add the condition to.
    """
    for condition in ahb_root.bedingungen:
        if condition.nummer == "2499":  # pragma: no cover
            raise ValueError(
                "Condition 2499 is used as a must-not pattern after sanitization of the AHB. "
                "Please check the AHB for correctness."
            )
    _set(
        ahb_root,
        "bedingungen",
        ahb_root.bedingungen
        + (
            ahb.Bedingung(
                nummer="2499",
                text='Ist immer falsch, um ein "Darf nicht" pattern umzusetzen: "X [2499]". '
                'Diese Expression wird für u.a. für Elemente genutzt, die laut MIG "unused" sind.',
            ),
        ),
    )


def remove_example_codes_recursive(
    element: (
        mig.MessageImplementationGuide
        | mig.SegmentGroup
        | ahb.Anwendungsfall
        | ahb.SegmentGroup
        | mig.Segment
        | ahb.Segment
        | mig.DataElementGroup
        | ahb.DataElementGroup
        | mig.DataElement
        | ahb.DataElement
    ),
) -> None:
    """
    Recursively removes example codes from MIG and AHB elements.

    :param element: The element to remove example codes from recursively.
    """
    match element:
        case mig.MessageImplementationGuide() | mig.SegmentGroup() | ahb.Anwendungsfall() | ahb.SegmentGroup():
            for sub_element in element.elements:
                remove_example_codes_recursive(sub_element)
        case mig.Segment() | ahb.Segment() | mig.DataElementGroup() | ahb.DataElementGroup():
            for sub_data_element in element.data_elements:
                remove_example_codes_recursive(sub_data_element)
        case mig.DataElement() | ahb.DataElement():
            # Remove example codes from MIG and AHB data elements
            object.__setattr__(
                element, "codes", tuple(code for code in element.codes if "Beispielcode" not in code.name)
            )
        case _:
            raise TypeError(f"Unexpected element type: {type(element)}. Expected MIG or AHB element types.")


def sanitize_ahb(mig_root: mig.MessageImplementationGuide, ahb_root: ahb.Anwendungshandbuch) -> None:
    """
    Sanitizes the AHB by adding unused MIG elements to the AHB.
    This function will add the following elements to the AHB:

    - Unused MIG data elements and data element groups to the AHB segment or data element group.
    - Unused MIG segments and segment groups to the AHB segment group or root.

    It will also add a condition to the AHB root element which is used to mark elements which are unused in the AHB
    and should not be provided in a message. The respective elements will have an ``ahb_status`` of ``X [2499]``.

    :param mig_root: The MIG root element to add unused elements from.
    :param ahb_root: The AHB root element to add unused elements to.
    """
    add_must_not_pattern_to_ahb_conditions(ahb_root)
    remove_example_codes_recursive(mig_root)
    for anwendungsfall in ahb_root.anwendungsfaelle:
        add_unused_segment_or_groups_to_ahb(mig_root, anwendungsfall)
        remove_example_codes_recursive(anwendungsfall)



"""Anwendungshandbuch SQL models"""

import uuid
from datetime import date
from typing import Optional, Union
from uuid import UUID

from efoli import EdifactFormat, EdifactFormatVersion

# pylint: disable=too-few-public-methods, duplicate-code, missing-function-docstring

# the structures are similar, still we decided against inheritance, so there's naturally a little bit of duplication


try:
    from sqlalchemy import CheckConstraint, UniqueConstraint
    from sqlmodel import Field, Relationship, SQLModel
except ImportError as import_error:
    import_error.msg += "; Did you install fundamend[sqlmodels] or did you try to import from fundamend.models instead?"
    # sqlmodel is only an optional dependency when fundamend is used to fill a database
    raise

from fundamend.models.anwendungshandbuch import Anwendungsfall as PydanticAnwendungsfall
from fundamend.models.anwendungshandbuch import Anwendungshandbuch as PydanticAnwendungshandbuch
from fundamend.models.anwendungshandbuch import Bedingung as PydanticBedingung
from fundamend.models.anwendungshandbuch import Code as PydanticCode
from fundamend.models.anwendungshandbuch import DataElement as PydanticDataElement
from fundamend.models.anwendungshandbuch import DataElementGroup as PydanticDataElementGroup
from fundamend.models.anwendungshandbuch import Paket as PydanticPaket
from fundamend.models.anwendungshandbuch import Segment as PydanticSegment
from fundamend.models.anwendungshandbuch import SegmentGroup as PydanticSegmentGroup
from fundamend.models.anwendungshandbuch import UbBedingung as PydanticUbBedingung


class Code(SQLModel, table=True):
    """
    A single code element inside an AHB DataElement, indicated by the `<Code>` tag.
    """

    __table_args__ = (
        UniqueConstraint("data_element_primary_key", "position", name="IX_position_once_per_data_element"),
    )
    primary_key: UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    name: str = Field(index=True)  # e.g. 'Netznutzungszeiten-Nachricht'
    description: str | None = Field(default=None, index=True)  # e.g. ''
    value: str | None = Field(default=None, index=True)  # e.g. 'UTILTS'
    ahb_status: str  #: e.g. 'X' # new for AHB
    position: Optional[int] = Field(default=None, index=True)

    dataelement: Union["DataElement", None] = Relationship(back_populates="codes")
    data_element_primary_key: UUID | None = Field(default=None, foreign_key="dataelement.primary_key")

    @classmethod
    def from_model(cls, model: PydanticCode, position: Optional[int] = None) -> "Code":
        return Code(
            name=model.name,
            description=model.description,
            value=model.value,
            ahb_status=model.ahb_status,
            position=position,
        )

    def to_model(self) -> PydanticCode:
        return PydanticCode(
            name=self.name,
            description=self.description,
            value=self.value,
            ahb_status=self.ahb_status,
        )


class DataElement(SQLModel, table=True):
    """
    A single data element, German 'Datenelement' inside an AHB Segment, indicated by the `<D_xxxx>` tag.
    This element can contain a single or multiple Code elements.
    """

    # Example:
    # <D_0065 Name="Nachrichtentyp-Kennung">
    #   <Code Name="Netznutzungszeiten-Nachricht" Description="" AHB_Status="X">UTILTS</Code>
    # </D_0065>
    primary_key: UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    id: str = Field(index=True)  # e.g. 'D_0065'
    name: str = Field(index=True)  # e.g. 'Nachrichtentyp-Kennung'
    codes: list[Code] = Relationship(back_populates="dataelement")
    position: Optional[int] = Field(default=None, index=True)
    ahb_status: Optional[str] = None
    dataelementgroup: Union["DataElementGroup", None] = Relationship(back_populates="data_elements")
    data_element_group_primary_key: UUID | None = Field(default=None, foreign_key="dataelementgroup.primary_key")

    segment: Union["Segment", None] = Relationship(back_populates="data_elements")
    segment_primary_key: UUID | None = Field(default=None, foreign_key="segment.primary_key")

    @classmethod
    def from_model(cls, model: PydanticDataElement, position: Optional[int] = None) -> "DataElement":
        result = DataElement(
            id=model.id,
            name=model.name,
            codes=[
                Code.from_model(pydantic_code, position=position_index)
                for position_index, pydantic_code in enumerate(model.codes)
            ],
            ahb_status=model.ahb_status,
            position=position,
        )
        return result

    def to_model(self) -> PydanticDataElement:
        return PydanticDataElement(
            id=self.id,
            name=self.name,
            ahb_status=self.ahb_status,
            codes=tuple(x.to_model() for x in sorted(self.codes, key=lambda y: y.position or 0)),
        )


class DataElementGroup(SQLModel, table=True):
    """
    A group of data elements, German 'Datenelementgruppe' inside the AHB, indicated by the `<C_xxxx>` tag.
    This model can contain both the 'Datenelement' and the 'Gruppendatenelement'
    """

    __table_args__ = (UniqueConstraint("segment_primary_key", "position", name="IX_position_once_per_segment"),)
    # "Die Datenelementgruppe C0829 enthält mehrere Gruppendatenelemente. Diese Datenelementgruppe enthält das
    # Gruppendatenelement DE3039, hier wird die MP-ID angegeben, sowie das DE3055, welches die code-vergebende Stelle
    # definiert. Diese Datenelementgruppe enthält des Weiteren das Gruppendatenelement DE113110, welches nur in der MIG
    # und nicht im AHB aufgeführt wird, um den Aufbau der EDIFACT Nachricht korrekt umsetzen zu können."
    # Quelle: Allgemeine Festlegungen Kapitel 6.1, Seite 52

    # Example:
    # <C_C002 Name="Dokumenten-/Nachrichtenname">
    #    <D_1001 Name="Dokumentenname, Code">
    #      <Code Name="Berechnungsformel" Description="" AHB_Status="X">Z36</Code>
    #   </D_1001>
    # </C_C002>
    primary_key: UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    id: str = Field(index=True)  # e.g. 'C_C082'
    name: str = Field(index=True)  # e.g. 'Dokumenten-/Nachrichtenname'
    data_elements: list[DataElement] = Relationship(back_populates="dataelementgroup")
    position: Optional[int] = Field(default=None, index=True)
    segment: Union["Segment", None] = Relationship(back_populates="data_element_groups")
    segment_primary_key: Union[UUID, None] = Field(default=None, foreign_key="segment.primary_key")

    @classmethod
    def from_model(cls, model: PydanticDataElementGroup, position: Optional[int] = None) -> "DataElementGroup":
        result = DataElementGroup(
            id=model.id,
            name=model.name,
            position=position,
        )
        for position_index, x in enumerate(model.data_elements):
            de = DataElement.from_model(x, position=position_index)
            result.data_elements.append(de)
        return result

    def to_model(self) -> PydanticDataElementGroup:
        return PydanticDataElementGroup(
            id=self.id,
            name=self.name,
            data_elements=tuple(x.to_model() for x in sorted(self.data_elements, key=lambda y: y.position or 0)),
        )


class Segment(SQLModel, table=True):
    """
    A segment inside an AHB, indicated by the `<S_xxxx>` tag.
    This model can contain both data elements and data element groups.
    """

    # Example:
    # pylint:disable=line-too-long
    #  <S_BGM Name="Beginn der Nachricht" Number="00002" AHB_Status="Muss">
    #    <C_C002 Name="Dokumenten-/Nachrichtenname">
    #       <D_1001 Name="Dokumentenname, Code">
    #         <Code Name="Berechnungsformel" Description="" AHB_Status="X">Z36</Code>
    #       </D_1001>
    #    </C_C002>
    #    <C_C106 Name="Dokumenten-/Nachrichten-Identifikation">
    #       <D_1004 Name="Dokumentennummer" AHB_Status="X"/>
    #    </C_C106>
    # </S_BGM>
    primary_key: UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    id: str = Field(index=True)  #: e.g. 'BGM'
    name: str = Field(index=True)  #: e.g. 'Beginn der Nachricht'
    number: str = Field(index=True)  #: e.g. '00002'
    ahb_status: str | None  #: e.g. 'Muss'
    is_on_uebertragungsdatei_level: bool
    data_elements: list[DataElement] = Relationship(back_populates="segment")
    data_element_groups: list[DataElementGroup] = Relationship(back_populates="segment")
    position: Optional[int] = Field(default=None, index=True)

    segmentgroup: Union["SegmentGroup", None] = Relationship(back_populates="segments")
    segmentgroup_primary_key: Union[UUID, None] = Field(default=None, foreign_key="segmentgroup.primary_key")

    anwendungsfall: Union["Anwendungsfall", None] = Relationship(back_populates="segments")
    anwendungsfall_primary_key: Union[UUID, None] = Field(default=None, foreign_key="anwendungsfall.primary_key")

    @classmethod
    def from_model(cls, model: PydanticSegment, position: Optional[int] = None) -> "Segment":
        result = Segment(
            id=model.id,
            name=model.name,
            number=model.number,
            ahb_status=model.ahb_status,
            position=position,
            is_on_uebertragungsdatei_level=model.is_on_uebertragungsdatei_level,
        )
        for _position, element in enumerate(model.data_elements):
            if isinstance(element, PydanticDataElement):
                result.data_elements.append(DataElement.from_model(element, position=_position))
                continue
            if isinstance(element, PydanticDataElementGroup):
                result.data_element_groups.append(DataElementGroup.from_model(element, position=_position))
                continue
        return result

    def to_model(self) -> PydanticSegment:
        return PydanticSegment(
            id=self.id,
            name=self.name,
            number=self.number,
            ahb_status=self.ahb_status,
            data_elements=tuple(
                x.to_model()
                for x in sorted(
                    (self.data_elements or []) + (self.data_element_groups or []),  # type:ignore[operator]
                    key=lambda y: y.position or 0,
                )
            ),
            is_on_uebertragungsdatei_level=self.is_on_uebertragungsdatei_level,
        )


class SegmentGroupLink(SQLModel, table=True):
    """artificial construct in SQL to model Segment Groups being nested in other Segment Groups"""

    parent_id: UUID | None = Field(default=None, foreign_key="segmentgroup.primary_key", primary_key=True)
    child_id: UUID | None = Field(default=None, foreign_key="segmentgroup.primary_key", primary_key=True)


class SegmentGroup(SQLModel, table=True):
    """
    A 'Segmentgruppe' inside an AHB, indicated by the `<G_xxxx>` tag.
    This model can contain both Segments and segment groups.
    """

    # Example:
    # <G_SG6 Name="Prüfidentifikator" AHB_Status="Muss">
    #   <S_RFF Name="Prüfidentifikator" Number="00012" AHB_Status="Muss">
    #     <C_C506 Name="Referenz">
    #       <D_1153 Name="Referenz, Qualifier">
    #         <Code Name="Prüfidentifikator" Description="" AHB_Status="X">Z13</Code>
    #       </D_1153>
    #       <D_1154 Name="Referenz, Identifikation">
    #         <Code Name="Berechnungsformel" Description="" AHB_Status="X">25001</Code>
    #       </D_1154>
    #     </C_C506>
    #   </S_RFF>
    #  </G_SG6>
    primary_key: UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    id: str = Field(index=True)  #: e.g. 'SG6'
    name: str = Field(index=True)  #: e.g. 'Prüfidentifikator'
    ahb_status: str | None  #: e.g. 'Muss'
    segments: list[Segment] = Relationship(back_populates="segmentgroup")
    position: Optional[int] = Field(default=None, index=True)
    # Define self-referential relationship
    segment_groups: list["SegmentGroup"] = Relationship(
        back_populates="parent_segment_group",
        sa_relationship_kwargs={
            "primaryjoin": "SegmentGroup.primary_key==SegmentGroupLink.parent_id",
            "secondaryjoin": "SegmentGroup.primary_key==SegmentGroupLink.child_id",
            "foreign_keys": "[SegmentGroupLink.parent_id, SegmentGroupLink.child_id]",
        },
        link_model=SegmentGroupLink,
    )

    parent_segment_group: Optional["SegmentGroup"] = Relationship(
        back_populates="segment_groups",
        sa_relationship_kwargs={
            "primaryjoin": "SegmentGroup.primary_key==SegmentGroupLink.child_id",
            "secondaryjoin": "SegmentGroup.primary_key==SegmentGroupLink.parent_id",
            "foreign_keys": "[SegmentGroupLink.child_id, SegmentGroupLink.parent_id]",
        },
        link_model=SegmentGroupLink,
    )

    anwendungsfall: Union["Anwendungsfall", None] = Relationship(back_populates="segment_groups")
    anwendungsfall_primary_key: Union[UUID, None] = Field(default=None, foreign_key="anwendungsfall.primary_key")

    @classmethod
    def from_model(cls, model: PydanticSegmentGroup, position: Optional[int] = None) -> "SegmentGroup":
        result = SegmentGroup(
            id=model.id,
            name=model.name,
            ahb_status=model.ahb_status,
            position=position,
        )
        for _position, element in enumerate(model.elements):
            if isinstance(element, PydanticSegment):
                result.segments.append(Segment.from_model(element, position=_position))
                continue
            if isinstance(element, PydanticSegmentGroup):
                sg = SegmentGroup.from_model(element, position=_position)
                if result.segment_groups is None:
                    result.segment_groups = [sg]
                else:
                    result.segment_groups.append(sg)
                continue
        return result

    def to_model(self) -> PydanticSegmentGroup:
        return PydanticSegmentGroup(
            id=self.id,
            name=self.name,
            ahb_status=self.ahb_status,
            elements=tuple(
                x.to_model()
                for x in sorted(
                    (self.segments or []) + (self.segment_groups or []),  # type:ignore[operator]
                    key=lambda y: y.position or 0,
                )
            ),
        )


class Anwendungsfall(SQLModel, table=True):
    """
    One 'Anwendungsfall', indicated by `<AWF>` tag, corresponds to one Prüfidentifikator or type of Message
    """

    __table_args__ = (UniqueConstraint("anwendungshandbuch_primary_key", "position", name="IX_position_once_per_ahb"),)
    # Example:
    # <AWF Pruefidentifikator="25001" Beschreibung="Berechnungsformel" Kommunikation_von="NB an MSB / LF">
    #   <M_UTILTS>
    #     <S_UNH Name="Nachrichten-Kopfsegment" Number="00001" AHB_Status="Muss">
    #    ...
    #     </S_UNT>
    #   </M_UTILTS>
    # </AWF>
    primary_key: UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    pruefidentifikator: str = Field(index=True)  #: e.g. '25001'
    beschreibung: str = Field(index=True)  #: e.g. 'Berechnungsformel'
    kommunikation_von: str  #: e.g. 'NB an MSB / LF'
    format: EdifactFormat = Field(index=True)  #: e.g. 'UTILTS'
    segments: list[Segment] = Relationship(back_populates="anwendungsfall")
    segment_groups: list[SegmentGroup] = Relationship(back_populates="anwendungsfall")
    position: Optional[int] = Field(default=None, index=True)
    anwendungshandbuch: Union["Anwendungshandbuch", None] = Relationship(back_populates="anwendungsfaelle")
    anwendungshandbuch_primary_key: Union[UUID, None] = Field(
        default=None, foreign_key="anwendungshandbuch.primary_key"
    )

    @classmethod
    def from_model(cls, model: PydanticAnwendungsfall, position: Optional[int] = None) -> "Anwendungsfall":
        result = Anwendungsfall(
            pruefidentifikator=model.pruefidentifikator,
            beschreibung=model.beschreibung,
            kommunikation_von=model.kommunikation_von,
            format=model.format,
            position=position,
        )
        for _position, element in enumerate(model.elements):
            if isinstance(element, PydanticSegment):
                result.segments.append(Segment.from_model(element, position=_position))
                continue
            if isinstance(element, PydanticSegmentGroup):
                result.segment_groups.append(SegmentGroup.from_model(element, position=_position))
                continue
        return result

    def to_model(self) -> PydanticAnwendungsfall:
        return PydanticAnwendungsfall(
            pruefidentifikator=self.pruefidentifikator,
            beschreibung=self.beschreibung,
            kommunikation_von=self.kommunikation_von,
            format=self.format,
            elements=tuple(
                x.to_model()
                for x in sorted(
                    ((self.segments or []) + (self.segment_groups or [])),  # type:ignore[operator]
                    key=lambda y: y.position or 0,
                )
            ),
        )


class Bedingung(SQLModel, table=True):
    """Ein ConditionKeyConditionText Mapping"""

    __table_args__ = (
        UniqueConstraint("anwendungshandbuch_primary_key", "position", name="IX_position_once_per_ahb"),
        UniqueConstraint("anwendungshandbuch_primary_key", "nummer", name="IX_bedingung_nummer_once_per_ahb"),
    )
    primary_key: UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    nummer: str = Field(index=True)  #: e.g. '1'
    text: str  #: e.g. 'Nur MP-ID aus Sparte Strom'
    position: Optional[int] = Field(default=None, index=True)
    anwendungshandbuch: Union["Anwendungshandbuch", None] = Relationship(back_populates="bedingungen")
    anwendungshandbuch_primary_key: Union[UUID, None] = Field(
        default=None, foreign_key="anwendungshandbuch.primary_key"
    )

    @classmethod
    def from_model(cls, model: PydanticBedingung, position: Optional[int] = None) -> "Bedingung":
        return Bedingung(nummer=model.nummer, text=model.text, position=position)

    def to_model(self) -> PydanticBedingung:
        return PydanticBedingung(nummer=self.nummer, text=self.text)


class UbBedingung(SQLModel, table=True):
    """Eine UB-Bedingung"""

    __table_args__ = (
        UniqueConstraint("anwendungshandbuch_primary_key", "position", name="IX_position_once_per_ahb"),
        UniqueConstraint("anwendungshandbuch_primary_key", "nummer", name="IX_ubbedingung_nummer_once_per_ahb"),
    )
    # Example:
    # <UB_Bedingung Nummer="[UB1]">([931] ∧ [932] [490]) ⊻ ([931] ∧ [933] [491])</UB_Bedingung>
    primary_key: UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    nummer: str = Field(index=True)  # e.g. 'UB1'
    text: str  #: e.g. '([931] ∧ [932] [490]) ⊻ ([931] ∧ [933] [491])'
    position: Optional[int] = Field(default=None, index=True)
    anwendungshandbuch: Union["Anwendungshandbuch", None] = Relationship(back_populates="ub_bedingungen")
    anwendungshandbuch_primary_key: Union[UUID, None] = Field(
        default=None, foreign_key="anwendungshandbuch.primary_key"
    )

    @classmethod
    def from_model(cls, model: PydanticUbBedingung, position: Optional[int] = None) -> "UbBedingung":
        return UbBedingung(nummer=model.nummer, text=model.text, position=position)

    def to_model(self) -> PydanticUbBedingung:
        return PydanticUbBedingung(nummer=self.nummer, text=self.text)


class Paket(SQLModel, table=True):
    """Ein Bedingungspaket/PackageKeyConditionText Mapping"""

    __table_args__ = (
        UniqueConstraint("anwendungshandbuch_primary_key", "position", name="IX_position_once_per_ahb"),
        UniqueConstraint("anwendungshandbuch_primary_key", "nummer", name="IX_paket_nummer_once_per_ahb"),
    )
    # Example:
    # <Paket Nummer="[1P]">--</Paket>
    primary_key: UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    nummer: str = Field(index=True)  #: e.g. '1P'
    text: str  #: e.g. '--'
    position: Optional[int] = Field(default=None, index=True)

    anwendungshandbuch: Union["Anwendungshandbuch", None] = Relationship(back_populates="pakete")
    anwendungshandbuch_primary_key: Union[UUID, None] = Field(
        default=None, foreign_key="anwendungshandbuch.primary_key"
    )

    @classmethod
    def from_model(cls, model: PydanticPaket, position: Optional[int] = None) -> "Paket":
        return Paket(nummer=model.nummer, text=model.text, position=position)

    def to_model(self) -> PydanticPaket:
        return PydanticPaket(nummer=self.nummer, text=self.text)


class Anwendungshandbuch(SQLModel, table=True):
    """
    Ein Anwendungshandbuch, indicated by the `<AHB` tag, bündelt verschiedene Nachrichtentypen/Anwendungsfälle im
    selben Format oder mit der selben regulatorischen Grundlage und stellt gemeinsame Pakete & Bedingungen bereit.
    """

    __table_args__ = (
        CheckConstraint("gueltig_bis IS NULL OR gueltig_bis > gueltig_von", name="gueltig_von_bis_sanity"),
    )
    primary_key: UUID = Field(primary_key=True, default_factory=uuid.uuid4)
    # Example:
    # <AHB Versionsnummer="1.1d" Veroeffentlichungsdatum="02.04.2024" Author="BDEW">
    #  <AWF Pruefidentifikator="25001" Beschreibung="Berechnungsformel" Kommunikation_von="NB an MSB / LF">
    #  </AWF>
    # </AHB>
    veroeffentlichungsdatum: date = Field(index=True)
    """publishing date"""

    autor: str
    """author, most likely 'BDEW'"""

    versionsnummer: str
    """e.g. '1.1d'"""
    anwendungsfaelle: list[Anwendungsfall] = Relationship(
        back_populates="anwendungshandbuch"
    )  #: die einzelnen Prüfidentifikatoren
    bedingungen: list[Bedingung] = Relationship(back_populates="anwendungshandbuch")
    ub_bedingungen: list[UbBedingung] = Relationship(back_populates="anwendungshandbuch")
    pakete: list[Paket] = Relationship(back_populates="anwendungshandbuch")

    # Die Gültig von/bis Datümer sind leider nicht teil des XML-Datenmodells, obwohl sie viel nützlicher wären als bspw.
    # das Veröffentlichungsdatum. Die Informationen darf man sich schön aus der mehr schlecht als recht gepflegten API
    # von bdew-mako.de rauskratzen. Sie sind aber nützlich um mehrere Versionen des AHBs in einer DB zu speichern.
    # Daher hier als SQLModel-Attribute ohne Entsprechung im XML/rohen Original-Datenmodell.
    gueltig_von: Optional[date] = Field(default=None, index=True)
    """
    inklusives Startdatum der Gültigkeit dieses AHBs (Deutsche Zeitzone)
    """
    gueltig_bis: Optional[date] = Field(default=None, index=True)
    """
    Ggf. exklusives Enddatum der Gültigkeit dieses AHBs (Deutsche Zeitzone).
    Wir verwenden None für ein offenes Ende, nicht 9999-12-31.
    """
    edifact_format_version: Optional[EdifactFormatVersion] = Field(default=None, index=True)
    """
    efoli format version (note that this is not derived from the gueltig von/bis dates but has to be set explicitly).
    It's also not a computed column although technically this might have been possible.
    For details about the type check the documentation of the EdifactFormatVersion enum from the efoli package.
    """

    @classmethod
    def from_model(cls, model: PydanticAnwendungshandbuch) -> "Anwendungshandbuch":
        return Anwendungshandbuch(
            veroeffentlichungsdatum=model.veroeffentlichungsdatum,
            autor=model.autor,
            versionsnummer=model.versionsnummer,
            bedingungen=[Bedingung.from_model(x, position=i) for i, x in enumerate(model.bedingungen)],
            ub_bedingungen=[UbBedingung.from_model(x, position=i) for i, x in enumerate(model.ub_bedingungen)],
            pakete=[Paket.from_model(x, position=i) for i, x in enumerate(model.pakete)],
            anwendungsfaelle=[
                Anwendungsfall.from_model(x, position=i)
                for i, x in enumerate(model.anwendungsfaelle)
                if not x.is_outdated
            ],
        )

    def to_model(self) -> PydanticAnwendungshandbuch:
        return PydanticAnwendungshandbuch(
            veroeffentlichungsdatum=self.veroeffentlichungsdatum,
            autor=self.autor,
            versionsnummer=self.versionsnummer,
            bedingungen=tuple(x.to_model() for x in sorted(self.bedingungen, key=lambda y: y.position or 0)),
            ub_bedingungen=tuple(x.to_model() for x in sorted(self.ub_bedingungen, key=lambda y: y.position or 0)),
            pakete=tuple(x.to_model() for x in sorted(self.pakete, key=lambda y: y.position or 0)),
            anwendungsfaelle=tuple(x.to_model() for x in sorted(self.anwendungsfaelle, key=lambda y: y.position or 0)),
        )


__all__ = ["Code", "DataElement", "DataElementGroup", "Segment", "SegmentGroup", "Anwendungsfall", "Bedingung"]



"""
models that work together with SQLModel+SQLAlchemy
you need to install fundamend[sqlmodels] to use this sub-package
"""

# the models here do NOT inherit from the original models, because I didn't manage to fix the issue that arise:
# <frozen abc>:106: in __new__
#     ???
# E   TypeError: Anwendungshandbuch.__init_subclass__() takes no keyword arguments
#
# or
#
# ..\.tox\dev\Lib\site-packages\sqlmodel\main.py:697: in get_sqlalchemy_type
#     raise ValueError(f"{type_} has no matching SQLAlchemy type")
# E   ValueError: <class 'fundamend.models.anwendungshandbuch.Anwendungshandbuch'> has no matching SQLAlchemy type
# => you need to keep the models in sync manually by now

from .ahbtabellen_view import AhbTabellenLine, create_ahbtabellen_view
from .ahbview import AhbHierarchyMaterialized, create_ahb_view, create_db_and_populate_with_ahb_view
from .anwendungshandbuch import (
    Anwendungsfall,
    Anwendungshandbuch,
    Bedingung,
    Code,
    DataElement,
    DataElementGroup,
    Segment,
    SegmentGroup,
)

__all__ = [
    "create_ahb_view",
    "AhbHierarchyMaterialized",
    "create_db_and_populate_with_ahb_view",
    "create_ahbtabellen_view",
    "AhbTabellenLine",
    "Code",
    "DataElement",
    "DataElementGroup",
    "Segment",
    "SegmentGroup",
    "Anwendungsfall",
    "Bedingung",
    "Anwendungshandbuch",
]



"""contains the entrypoint for the command line interface"""

from fundamend.commands import app

__all__ = ["app"]


def main() -> None:
    """entry point of the script defined in pyproject.toml"""
    app()


if __name__ == "__main__":
    app()
